/**
 * Permite que o usuário selecione uma seção da imagem
 * e busca por esse pedaço em outra imagem
 * Uso: correspond <imgInicial> <imgFinal>
 * 
 * @author Leandro Facchinetti
 * 
 * Parte da seleção da caixa baseada no
 * exemplo 4-1 do livro OpenCV Computer Vision
 * 
 * @todo nomalizar imagem, aplicar thresold
 */
#include <stdio.h>
#include <cv.h>
#include <highgui.h>
#include <time.h>

//#define DEBUG

/**
 * Se estiver definida esta constante, o programa faz
 * análise de se desempenho, escrevendo num arquivo de texto
 * chamado benchmark.txt. Nele sai o tempo e o resultado do programa
 */
#define BENCHMARK

/**
 * Se estiver setado, chama o algoritmo piramidal
 */
//#define USE_PYR

/**
 * O status da caixa pode ser: inativo, desenhando ou desenhado
 */
#define BOX_STATUS_SLEEP 0
#define BOX_STATUS_DRAWNING 1
#define BOX_STATUS_DRAWN 2

/**
 * Distância máxima entre dois histogramas
 * Para mudar o método de comparação entre histogramas
 * é preciso mudar esse valor
 */
#define MAX_DIST 0
 
/**
 * Callback instalado no mouse para que ele desenhe a
 * caixa de seleção sobre a imagem original
 * 
 * Permite que o usuário selecione uma caixa sobre a
 * imagem original. Ela será a área de interesse na
 * qual se buscará correlação.
 * É uma função callback instalada no mouse
 * É setada a variável global:
 * int boxStatus status da caixa: inativo, desenhando ou desenhado
 * (usada por main)
 * Normalmente o estado é inativo
 * Ao clicar, é mudado para desenhando
 * Logo que o usuário solta o mouse, o status é mudado para desenhado
 * 
 * @param event qual o evento do mouse que chamou
 * @param x coordenada no mouse no eixo x
 * @param y coordenada no mouse no eixo y
 * @param flags flags relacionadas ao evento do mouse
 * @param param retângulo de interesse
 */
void
mouseCallbackDrawBox( int event, int x, int y, int flags, void* param );

/**
 * Desenha um retângulo na imagem
 * 
 * Encapsula a rotina de desenho de retângulo
 * 
 * @param img imagem na qual será desenhada
 * @param rect coordenadas do retângulo que será desenhado
 */
void
drawBox( IplImage* img, CvRect rect );


/**
 * Procura pela área de correspondência entre imagens
 * 
 * Procura na ROI de imgFinal pela região do tamanho da ROI
 * de imgInicial que tenha a menor distância.
 * A distância é calculada pela comparação entre os histogramas.
 * Caso a ROI de imgInicial seja maior que a ROI de imgFinal,
 * é retornado um retângulo com medidas CvRect( -1, -1, 0, 0 )
 * Assume-se que as imagens tenham ambas 3 canais ou 1 canal
 * 
 * @param imgInicial imagem que contém o elemento buscado
 * @param imgFinal imagem em que se busca correlação
 * @return retângulo onde ocorreu a menor distância
 */
CvRect
correspond( IplImage *imgInicial, IplImage *imgFinal );

/**
 * Procura pela área de correspondência entre imagens,
 * usando um algoritmo piramidal
 * 
 * Procura na ROI de imgFinal pela região do tamanho da ROI
 * de imgInicial que tenha a menor distância.
 * A distância é calculada pela comparação entre os histogramas.
 * Caso a ROI de imgInicial seja maior que a ROI de imgFinal,
 * é retornado um retângulo com medidas CvRect( -1, -1, 0, 0 )
 * Assume-se que as imagens tenham ambas 3 canais ou 1 canal.
 * O algoritmo usa a correspond(), para cada etapa da pirâmide.
 * A ideia é fazer uma pirâmide de Gauss, para otimizar o
 * funcionamento da correspond(). A correspondência é procurada
 * na imagem reduzida e essa informação é usada para restringir a
 * busca na imagem maior.
 * Para poder redimensionar a imagem e fazer uso da característica
 * piramidal do algoritmo, é preciso que o tamanho da imagem seja
 * múltiplo de 2^reducao, sendo reducao a quantidade de vezes que a imagem será
 * reduzida
 * 
 * @param imgInicial imagem que contém o elemento buscado
 * @param imgFinal imagem em que se busca correlação
 * @param reducao quantidade de vezes que a imagem pode ser reduzida
 * quanto maior esse número, mais rápido e menos preciso é o algoritmo
 * @param janela tamanho da janela de erro quando é analisada a imagem
 * em resolução maior. Quanto maior esse número mais lento e preciso
 * é o algoritmo. A janela mínima é de 5px.
 * @return retângulo onde ocorreu a menor distância
 */
CvRect
pyrCorrespond( IplImage *imgInicialOriginal, IplImage *imgFinalOriginal, int reducao, int janela );
 
/**
 * Status da caixa: inativo, desenhando ou desenhado
 */
int boxStatus = BOX_STATUS_SLEEP;
 
void
drawBox( IplImage* img, CvRect rect )
{
	cvRectangle(
		img, 
		cvPoint( rect.x, rect.y ),
		cvPoint( rect.x + rect.width, rect.y + rect.height ),
		cvScalar( 0x00, 0x00, 0xff)
	);
}
 
int
main( int argc, char* argv[] )
{

#ifdef BENCHMARK
	// arquivo no qual será feita a saída da análise de
	// desempenho e a variável que guarda o tempo inicial,
	// antes da execução do algoritmo.
	FILE *fileBenchmark;
	clock_t benchmarkInitTime;
#endif

	// Imagem inicial, onde o usuário seleciona a região
	// de interesse
	IplImage *imgInicial;

	// Imagem final, onde se busca a correlação
	// de interesse
	IplImage *imgFinal;

	// imagem temporária usada para
	// o usuário desenhar sobre
	IplImage *tempInicial;

	// imagem temporária usada para
	// mostrar o resultado da correlação
	IplImage *tempFinal;

	// retângulo de interesse na imagem inicial
	CvRect box = cvRect( -1, -1, 0, 0 ),
		resposta;

	char imgInicialNome[256],
		imgFinalNome[256],
		peb = 0,
		verbose = 0;

	// checando se o programa foi chamado corretamente:
	// correspond <imagemInicial> <imagemFinal> [peb]
	//   [coordenadaX coordenadaY tamanhoH tamanhoV] [v]
	// peb se for para tratar as imagens em preto e branco
	// e as coordenadas do retângulo a ser buscado na imagem
	// destino. Se não for colocado, será aberta janela
	// para o usuário escolher a área com o mouse.
	// O parâmetro v determina que a saída não será numa janela,
	// mostrada para o usuário, com a imagem final e o retângulo
	// encontrado, mas sim impressa no console. É útil para batch
	// de automação de testes.
	//
	// esse switch resgata os valores passados por parâmetros e
	// encerra caso ele tenha tido chamado com parâmetros inválidos.
	// Os nomes das imagens inicial e final são resgatados em todos
	// os casos, por isso é feito depois, fora do switch
	switch( argc )
	{
		case 3:
		break;
		case 4:
			if( strcmp( argv[3], "peb" ) == 0 )
				peb = 1;
			if( strcmp( argv[3], "v" ) == 0 )
				verbose = 1;
		break;
		case 5:
			peb = 1;
			verbose = 1;
		break;
		case 7:
			box.x = atoi( argv[3] );
			box.y = atoi( argv[4] );
			box.width = atoi( argv[5] );
			box.height = atoi( argv[6] );
		break;
		case 8:
			if( strcmp( argv[3], "peb" ) == 0 )
			{
				peb = 1;
				box.x = atoi( argv[4] );
				box.y = atoi( argv[5] );
				box.width = atoi( argv[6] );
				box.height = atoi( argv[7] );
			}
			else
			{
				verbose = 1;
				box.x = atoi( argv[3] );
				box.y = atoi( argv[4] );
				box.width = atoi( argv[5] );
				box.height = atoi( argv[6] );
			}
		break;
		case 9:
			peb = 1;
			verbose = 1;
			box.x = atoi( argv[4] );
			box.y = atoi( argv[5] );
			box.width = atoi( argv[6] );
			box.height = atoi( argv[7] );
		break;
		default:
		printf( "correspond <imagemInicial> <imagemFinal> [peb]\
 [coordenadaX coordenadaY tamanhoH tamanhoV] [v]" );
		exit( 1 );
	}

	// resgatando os nomes das imagens inicial e final
	strcpy( imgInicialNome, argv[1] );
	strcpy( imgFinalNome, argv[2] );

	if( ! verbose )
		printf( "Correspondencia entre imagens (correlacao)\n\
para sair, digite <esc>\n" );
	// iniciando as variáveis, carregando as imagens
	// e iniciando as janelas
	if( peb )
	{
		imgInicial = cvLoadImage( 
			imgInicialNome,
			CV_LOAD_IMAGE_GRAYSCALE
		);
		imgFinal = cvLoadImage( 
			imgFinalNome,
			CV_LOAD_IMAGE_GRAYSCALE
		);
	}
	else
	{
		imgInicial = cvLoadImage( 
			imgInicialNome,
			CV_LOAD_IMAGE_COLOR
		);
		imgFinal = cvLoadImage( 
			imgFinalNome,
			CV_LOAD_IMAGE_COLOR
		);
	}

	tempInicial = cvCloneImage( imgInicial );
	tempFinal = cvCloneImage( imgFinal );

	if( ! verbose )
		cvNamedWindow( "Final", CV_WINDOW_AUTOSIZE );
	// se o usuário não setou uma caixa como parâmetro
	if( box.x == -1 )
	{
		cvNamedWindow( "Selecione uma regiao da imagem", CV_WINDOW_AUTOSIZE );

		// instalando o callback
		cvSetMouseCallback( 
			"Selecione uma regiao da imagem",
			mouseCallbackDrawBox, 
			(void*) &box
		);

		// Permitindo que o usuário selecione uma região da imagem
		// É criada uma cópia da imagem original, para que seja
		// possível desenhar o retângulo nela.
		// Caso ele tenha acabado de desenhar, é chamada a
		// rotina de correspondência e o resultado é mostrado
		// Além disso, espera-se por 15 milissegundos
		// por um <esc> do usuário, que termina o programa
		while( cvWaitKey(15) != 27 )
		{
			if( boxStatus == BOX_STATUS_DRAWNING )
			{
				cvCopyImage( imgInicial, tempInicial );
				drawBox( tempInicial, box );
			}
			if( boxStatus == BOX_STATUS_DRAWN )
			{
				boxStatus = BOX_STATUS_SLEEP;
				cvCopyImage( imgFinal, tempFinal );
				cvSetImageROI( imgInicial, box );
#ifdef BENCHMARK
				benchmarkInitTime = clock();
#endif
#ifdef USE_PYR
				resposta = pyrCorrespond( imgInicial, imgFinal, 3, 30 );
#else
				resposta = correspond( imgInicial, imgFinal );
#endif
#ifdef BENCHMARK
				fileBenchmark = fopen( "benchmark.txt", "a" );
				CvRect iniROI = cvGetImageROI( imgInicial );
#ifdef USE_PYR
				fprintf( fileBenchmark, "\tTempo execução do algoritmo (piramidal): \
%lf\t[(%d, %d, %d, %d) -> (%d, %d, %d, %d)]\n",
					(double)( clock() - benchmarkInitTime ) / CLOCKS_PER_SEC,
					iniROI.x, iniROI.y, iniROI.width, iniROI.height,
					resposta.x, resposta.y, resposta.width, resposta.height );
#else
				fprintf( fileBenchmark, "\tTempo execução do algoritmo (normal): \
%lf\t[(%d, %d, %d, %d) -> (%d, %d, %d, %d)]\n",
					(double)( clock() - benchmarkInitTime ) / CLOCKS_PER_SEC,
					iniROI.x, iniROI.y, iniROI.width, iniROI.height,
					resposta.x, resposta.y, resposta.width, resposta.height );
#endif
				fclose( fileBenchmark );
#endif
				if( verbose )
					printf( "%d %d %d %d\n", resposta.x, resposta.y,
						resposta.width, resposta.height );
				else
					drawBox(
						tempFinal,
						resposta
					);
				cvResetImageROI( imgInicial );
			}
			cvShowImage( "Selecione uma regiao da imagem", tempInicial );
			
			if( ! verbose )
				cvShowImage( "Final", tempFinal );
		}
	}
	// se o usuário setou uma caixa como parâmetro
	else
	{
		cvSetImageROI( imgInicial, box );
		resposta = correspond( imgInicial, imgFinal );
		if( verbose )
			printf( "%d %d %d %d\n", resposta.x, resposta.y,
				resposta.width, resposta.height );
		else
		{
			drawBox(
				tempFinal,
				resposta
			);
			cvShowImage( "Final", tempFinal );
			cvWaitKey( 0 );
		}
	}

	// sendo limpinho
	cvReleaseImage( &imgInicial );
	cvReleaseImage( &imgFinal );
	cvReleaseImage( &tempInicial );
	cvReleaseImage( &tempFinal );
	if( box.x == -1 )
		cvDestroyWindow( "Selecione uma regiao da imagem" );
	if( ! verbose )
		cvDestroyWindow( "Final" );
}
 
void
mouseCallbackDrawBox( int event, int x, int y, int flags, void* param )
{
	// convertendo o parâmetro para um ponteiro do tipo certo CvRect
	CvRect *box = (CvRect*) param;

	// Quando o mouse é pressionado, começa a desenhar.
	// Quando ele se move, a caixa é redimensionada
	// Quando o mouse é solto, é acertada a caixa
	switch( event )
	{
		case CV_EVENT_MOUSEMOVE:
		{
			if( boxStatus == BOX_STATUS_DRAWNING )
			{
				box->width  = x - box->x;
				box->height = y - box->y;
			}
		}
		break;
		case CV_EVENT_LBUTTONDOWN:
		{
			boxStatus = BOX_STATUS_DRAWNING;
			box->x = x;
			box->y = y;
			box->width = 0;
			box->height = 0;
		}
		break;   
		case CV_EVENT_LBUTTONUP:
		{
			boxStatus = BOX_STATUS_DRAWN;
			if( box->width < 0 )
			{ 
				box->x += box->width;  
				box->width *= -1; 
			}
			if( box->height < 0 )
			{
				box->y += box->height; 
				box->height *= -1; 
			}
		}
		break;   
	}
}

CvRect
correspond( IplImage *imgInicial, IplImage *imgFinal )
{
	CvRect box = cvGetImageROI( imgInicial );

#ifdef DEBUG
	printf(
		"\nAplicando a fc de correlacao c caixa = \
CvRect( %d, %d, %d, %d )",
		box.x,
		box.y,
		box.width,
		box.height
	);
#endif

	// se as imagens só tem um canal, é usado o
	// canal R para fazer os cálculos
	// e as outras estruturas são alocadas à toa

	int peb = (imgInicial->nChannels == 1);

	// cada canal da imagem inicial
	IplImage *iniR = cvCreateImage(
		cvGetSize( imgInicial ),
		IPL_DEPTH_8U,
		1
	);
	IplImage *iniG = cvCreateImage(
		cvGetSize( imgInicial ),
		IPL_DEPTH_8U,
		1
	);
	IplImage *iniB = cvCreateImage(
		cvGetSize( imgInicial ),
		IPL_DEPTH_8U,
		1
	);
	// vetor dos canais da imagem inicial
	IplImage *iniPlanes[] =  { iniR, iniG, iniB };
	// cada canal da imagem final
	IplImage *finR = cvCreateImage(
		cvGetSize( imgFinal ),
		IPL_DEPTH_8U,
		1
	);
	IplImage *finG = cvCreateImage(
		cvGetSize( imgFinal ),
		IPL_DEPTH_8U,
		1
	);
	IplImage *finB = cvCreateImage(
		cvGetSize( imgFinal ),
		IPL_DEPTH_8U,
		1
	);
	// vetor dos canais da imagem inicial
	IplImage *finPlanes[] = { finR, finG, finB };
	// ROI da imagem inicial
	CvRect iniROI = cvGetImageROI( imgInicial ),
		// ROI da imagem final
		finROI = cvGetImageROI( imgFinal ),
		// janela onde ocorre mínima distância,
		// é o retorno da rotina
		minWindow = cvRect( -1, -1, 0, 0 ),
		// janela que percorrerá a imgFinal
		window;
	// quantidade de "cestos" por canal da imagem
	int bins[] = { 30, 30, 30 };
	// faixas por canal (RGB == [0,255])
	float rRanges[] = { 0, 255 };
	float gRanges[] = { 0, 255 };
	float bRanges[] = { 0, 255 };
	float *ranges[] = { rRanges, gRanges, bRanges };
	// histograma das imagens
	CvHistogram *iniHist,
				*finHist;
	if( peb )
	{
		iniHist = cvCreateHist(
			1,
			bins,
			CV_HIST_ARRAY,
			ranges,
			1
		);
		finHist = cvCreateHist(
			1,
			bins,
			CV_HIST_ARRAY,
			ranges,
			1
		);
	}
	else
	{
		iniHist = cvCreateHist(
			3,
			bins,
			CV_HIST_ARRAY,
			ranges,
			1
		);
		finHist = cvCreateHist(
			3,
			bins,
			CV_HIST_ARRAY,
			ranges,
			1
		);
	}
	// distância mínima entre histogramas
	// deve estar com o valor de maior disparidade
	double minDist = MAX_DIST,
	// distância entre histogramas
		dist;

	// testar se as ROIs são válidas
	if( ( iniROI.width < finROI.width ) && ( iniROI.height < finROI.height ) )
	{
		// decompondo as imagens em canais
		if( peb )
		{
			iniR = cvCloneImage( imgInicial );
			finR = cvCloneImage( imgFinal );
		}
		else
		{
			cvSplit( imgInicial, iniR, iniG, iniB, NULL );
			cvSplit( imgFinal, finR, finG, finB, NULL );
		}

		// calculando o histograma da imagem inicial
		if( peb )
			cvCalcHist( &iniR, iniHist, 0, 0 );
		else
			cvCalcHist( iniPlanes, iniHist, 0, 0 );
		cvNormalizeHist( iniHist, 1 );
		// percorrendo com a janela a imagem final
		window = cvRect( 0, 0, iniROI.width, iniROI.height );
		for( window.y = 0; window.y < finROI.height - iniROI.height; window.y += 1 )
			for( window.x = 0; window.x < finROI.width - iniROI.width; window.x += 1 )
			{
				// colocando a janela no lugar calculando o histograma nela
				if( peb )
				{
					cvSetImageROI( finR, window );
					cvCalcHist( &finR, finHist, 0, 0 );
				}
				else
				{
					cvSetImageROI( finR, window );
					cvSetImageROI( finG, window );
					cvSetImageROI( finB, window );
					cvCalcHist( finPlanes, finHist, 0, 0 );
				}

				// comparando os histogramas
				cvNormalizeHist( finHist, 1 );
				dist = cvCompareHist( iniHist, finHist, CV_COMP_INTERSECT );
				if( dist > minDist )
				{
					minDist = dist;
					minWindow = cvRect(
						finROI.x + window.x,
						finROI.y + window.y,
						iniROI.width,
						iniROI.height
					);
				}
			}
	}

	// limpando a bagunça
	cvReleaseImage( &iniR );
	cvReleaseImage( &iniG );
	cvReleaseImage( &iniB );
	cvReleaseImage( &finR );
	cvReleaseImage( &finG );
	cvReleaseImage( &finB );
	cvReleaseHist( &iniHist );
	cvReleaseHist( &finHist );

#ifdef DEBUG
	printf( "\nNem demorou tanto!! :P" );
#endif
	return minWindow;
}

CvRect
pyrCorrespond( IplImage *imgInicialOriginal, IplImage *imgFinalOriginal, int reducao, int janela )
{
	// receberá o resultado de correspond() com os valores
	// apropriados. É o resultado da rotina
	CvRect resp;

	// trabalho com copia da imagem porque senão buga quando executa pela segunda
	// vez.
	IplImage *imgInicial = cvCloneImage( imgInicialOriginal );
	IplImage *imgFinal = cvCloneImage( imgFinalOriginal );

	//garantindo uma janela mínima de 5px.
	if ( janela < 5 )
		janela = 5;

	// verificando se a imagem tem dimensões múltiplas de 2,
	// caso contrário não é possível fazer a pirâmide Gaussiana
	CvRect ROIImgInicial = cvGetImageROI( imgInicial );
	CvRect ROIImgFinal = cvGetImageROI( imgFinal );
	// para conseguir resgatar o tamanho da imagem, temos que degradar
	// as imagens originais. Retira-se seu ROI
	// É preciso reconstruí-las, para usar!!
	// É isso que o if a seguir faz!
	cvResetImageROI( imgInicial );
	cvResetImageROI( imgFinal );
	CvSize sizeImgInicial = cvGetSize( imgInicial );
	CvSize sizeImgFinal = cvGetSize( imgFinal );

	if( ( reducao > 0 ) &&
		( sizeImgInicial.height % 2 == 0 ) &&
		( sizeImgInicial.width % 2 == 0 )  )
	{
		IplImage *pyrImgInicial = cvCreateImage(
			cvSize( sizeImgInicial.width / 2, sizeImgInicial.height / 2 ),
			imgInicial->depth,
			imgInicial->nChannels
		);
		IplImage *pyrImgFinal = cvCreateImage(
			cvSize( sizeImgFinal.width / 2, sizeImgFinal.height / 2 ),
			imgFinal->depth,
			imgFinal->nChannels
		);
		
		CvRect ROIPyrImgInicial = cvRect( ROIImgInicial.x / 2,
			ROIImgInicial.y / 2,
			ROIImgInicial.width / 2,
			ROIImgInicial.height / 2 );
		CvRect ROIPyrImgFinal = cvRect( ROIImgFinal.x / 2,
			ROIImgFinal.y / 2,
			ROIImgFinal.width / 2,
			ROIImgFinal.height / 2 );

		cvPyrDown( imgInicial, pyrImgInicial );
		cvPyrDown( imgFinal, pyrImgFinal );

#ifdef DEBUG
		cvNamedWindow( "DEBUG imgInicial", CV_WINDOW_AUTOSIZE );
		cvNamedWindow( "DEBUG imgFinal", CV_WINDOW_AUTOSIZE );
		cvNamedWindow( "DEBUG pyrImgInicial", CV_WINDOW_AUTOSIZE );
		cvNamedWindow( "DEBUG pyrImgFinal", CV_WINDOW_AUTOSIZE );

		cvShowImage( "DEBUG imgInicial", imgInicial );
		cvShowImage( "DEBUG imgFinal", imgFinal );
		cvShowImage( "DEBUG pyrImgInicial", pyrImgInicial );
		cvShowImage( "DEBUG pyrImgFinal", pyrImgFinal );

		cvWaitKey( 0 );

		cvDestroyWindow( "DEBUG imgInicial" );
		cvDestroyWindow( "DEBUG imgFinal" );
		cvDestroyWindow( "DEBUG pyrImgInicial" );
		cvDestroyWindow( "DEBUG pyrImgFinal" );
#endif

		cvSetImageROI( pyrImgInicial, ROIPyrImgInicial );
		cvSetImageROI( pyrImgFinal, ROIPyrImgFinal );
		
		resp = pyrCorrespond( pyrImgInicial, pyrImgFinal, reducao - 1, janela / 2 );
		resp.x = ( resp.x * 2 - janela > 0 ) ? resp.x * 2 - 10 : 0;
		resp.y = ( resp.y * 2 - janela > 0 ) ? resp.y * 2 - 10 : 0;
		resp.width = ( resp.width * 2 + janela < imgFinal->width ) ?
			resp.width * 2 + janela : imgFinal->width;
		resp.height = ( resp.height * 2 + janela < imgFinal->height ) ?
			resp.height * 2 + janela : imgFinal->height;

		cvSetImageROI( imgInicial, ROIImgInicial );
		cvSetImageROI( imgFinal, resp );
		
		// limpando a bagunça
		cvReleaseImage( &pyrImgInicial );
		cvReleaseImage( &pyrImgFinal );
	}
	else
	{
		cvSetImageROI( imgInicial, ROIImgInicial );
		cvSetImageROI( imgFinal, ROIImgFinal );
	}
	
	resp = correspond( imgInicial, imgFinal );

	return resp;
}
